import os

import numpy as np
import onnx

from watermark_verify import logger
from watermark_verify.tools import secret_label_func


def label_verification(model_filename: str) -> bool:
    """
    模型标签提取验证
    :param model_filename: 模型权重文件，om格式
    :return: 模型标签验证结果
    """
    root_dir = os.path.dirname(model_filename)
    label_check_result = False
    logger.info(f"开始检测模型水印, model_filename: {model_filename}, root_dir: {root_dir}")
    # step 1 获取签名公钥信息，检查投影矩阵位置
    public_key_txt = os.path.join(root_dir, 'keys', 'public.key')
    x_random_file = os.path.join(root_dir, 'keys', 'key.npy')
    if not os.path.exists(x_random_file):
        logger.error(f"x_random_file={x_random_file}, 投影矩阵保存文件不存在")
    if not os.path.exists(public_key_txt):
        logger.error(f"public_key_txt={public_key_txt}, 签名公钥文件不存在")
        raise FileExistsError("签名公钥文件不存在")
    with open(public_key_txt, 'r') as file:
        public_key = file.read()
    logger.debug(f"x_random_file={x_random_file}, public_key_txt={public_key_txt}, public_key={public_key}")
    if not public_key or public_key == '':
        logger.error(f"获取的签名公钥信息为空, public_key={public_key}")
        raise RuntimeError("获取的签名公钥信息为空")

    # step 2 获取模型指定嵌入层权重，根据白盒水印算法，计算水印，并转化为字符串
    model = onnx.load(model_filename)  # 加载 ONNX 模型
    graph = model.graph  # 获取模型图（graph）
    weights = []
    # 遍历图中的节点
    for node in graph.node:
        if node.op_type == "Conv":  # 查找嵌入白盒水印的卷积层节点，卷积层名字可解析onnx文件后查找得到
            weight_name = node.input[1]  # 通常第一个是输入x、第二个输入是权重w、第三个是偏置b
            for initializer in graph.initializer:
                if initializer.name == weight_name:
                    # 获取权重数据
                    weights.append(onnx.numpy_helper.to_array(initializer))
                    logger.debug(f"Weight shape for {node.name}: {weights[len(weights) - 1].shape}")
    weights = weights[9:13]
    weights = [np.transpose(weight, (2, 3, 1, 0)) for weight in weights]  # 将onnx文件的权重格式由(out_channels, in_channels, kernel_height, kernel_width)，转换为(kernel_height, kernel_width, in_channels, out_channels)
    x_random = np.load(x_random_file)
    # 打印权重和投影矩阵
    logger.debug(f"Weights: {weights}")
    logger.debug(f"x_random: {x_random}")
    # 计算嵌入的白盒水印
    w = flatten_parameters(weights)
    prob = get_prob(x_random, w)
    decode = np.where(prob > 0.5, 1, 0)
    code_string = ''.join([str(x) for x in decode.tolist()])
    secret_label = bin2string(code_string)
    logger.debug(f"extract secret_label: {secret_label}")
    # step 3 验证密码标签
    label_check_result = secret_label_func.verify_secret_label(secret_label=secret_label, public_key=public_key)
    return label_check_result


def flatten_parameters(weights):
    """
    处理传入的卷积层的权重参数 (基于 NumPy 实现)

    :param weights: 指定卷积层的权重列表 (NumPy 数组)
    :return: 处理完成返回的 NumPy 数组
    """
    return np.concatenate([np.mean(x, axis=3).reshape(-1) for x in weights])


def get_prob(x_random: np.ndarray, w: np.ndarray) -> np.ndarray:
    """
    获取投影矩阵与权重向量的计算结果 (基于 NumPy 实现)

    :param x_random: 投影矩阵 (NumPy 数组)
    :param w: 权重向量 (NumPy 数组)
    :return: 计算结果 (NumPy 数组)
    """
    # 进行矩阵乘法
    mm = np.dot(x_random, w.reshape((w.shape[0], 1)))

    # 计算 Sigmoid 函数
    sigmoid_mm = 1 / (1 + np.exp(-mm))

    # 将结果展平
    return sigmoid_mm.flatten()


def bin2string(binary_string):
    return ''.join(chr(int(binary_string[i:i + 8], 2)) for i in range(0, len(binary_string), 8))
